/* finetd, an Internet Superserver
 *
 * Supports IPv4, IPv6 and UNIX domain sockets
 *
 ** Kris Katterjohn <katterjohn@gmail.com>
 ** http://finetd.sourceforge.net
 *
 * Copyright (c) 2008 Kris Katterjohn
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* $Id: parseconf.c 11 2008-02-22 18:43:06Z kjak $ */
/* https://finetd.svn.sourceforge.net/svnroot/finetd */

#include <unistd.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <limits.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>

#include "finetd.h"
#include "builtin.h"
#include "error.h"
#include "mem.h"

/* Takes a string like
 *      "/usr/sbin/tcpd /usr/sbin/ftpd [...]"
 * and creates an array like
 *      { "/usr/sbin/tcpd", "/usr/sbin/ftpd", ..., NULL }
 */
static char **makeargv(char *s)
{
	int i, k;
	int len = strlen(s);
	char *c;
	char **a;

	for (i = 0, k = 1; i < len; i++) {
		if (isspace(s[i])) {
			s[i] = 0;
		} else {
			if (s[i - 1] == 0)
				k++;
		}
	}

	a = Malloc(sizeof(char *) * (k + 1));

	for (c = s, i = 0; i < k; i++) {
		a[i] = strdup(c);

		for (c += strlen(c); *c == 0; c++);
	}

	a[i] = NULL;

	return a;
}

/* Things like invalid protocol or service names are handled in parseconf(),
 * but things that are based on multiple fields in an entry are tested here
 */
static void verifyentry(struct sserv *s, int lineno)
{
	if (!s->internal && access(s->path, X_OK) == -1)
		die("%s (line %d): cannot execute path \"%s\"", s->name, lineno, s->path);

	if (s->internal && !is_builtin(s->port, s->proto))
		die("%s (line %d): service isn't built-in!", s->name, lineno);

	if (s->type == SOCK_DGRAM && !s->wait)
		die("%s (line %d): datagram services require 'wait'", s->name, lineno);

	if (s->proto == IPPROTO_TCP && s->type != SOCK_STREAM)
		die("%s (line %d): can't use tcp with non-stream socket!", s->name, lineno);

	if (s->proto == IPPROTO_UDP && s->type != SOCK_DGRAM)
		die("%s (line %d): can't use udp with non-dgram socket!", s->name, lineno);
}

/* There are two possibilities for a valid entry:
 *
 *    {name} {type} {proto} [no]wait {user[:{group}]} {path} {args}
 *
 *    {name} {type} {proto} [no]wait {user[:{group}]} internal
 *
 * The first one causes an exec of the specified process while the second
 * uses one of the builtin_* functions.
 *
 * UNIX sockets can be prefixed with the following information:
 *
 *    :{user}:{group}:{mode}:/path/to/sock
 *
 * Comments start with a #
 */
struct sserv *parseconf(char *file)
{
	char *c, *x;
	char name[65];
	char type[10];
	char proto[16];
	char wait[7];
	char user[65];
	char path[PATH_MAX];
	char line[1024];
	int nline;
	int nservs;
	int i;
	FILE *fp;
	struct passwd *pw;
	struct group *gr;
	struct servent *port;
	struct sserv *sservs;
	struct sserv s;

	if (!(fp = fopen(file, "r")))
		pdie("fopen(\"%s\")", file);

	nline = 0;

	/* at least one for NULL */
	nservs = 1;
	sservs = Malloc(sizeof(struct sserv) * nservs);

	while (fgets(line, sizeof line, fp)) {
		nline++;

		if ((c = strchr(line, '#')))
			*c = 0;

		if ((c = strchr(line, '\n')))
			*c = 0;

		c = line;

		for (; *c == ' ' || *c == '\t'; c++);

		if (!*c)
			continue;

		memset(&s, 0, sizeof s);

		memset(name, 0, sizeof name);
		memset(type, 0, sizeof type);
		memset(proto, 0, sizeof proto);
		memset(wait, 0, sizeof wait);
		memset(user, 0, sizeof user);
		memset(path, 0, sizeof path);
		i = 0;

		sscanf(line, "%64s %9s %15s %6s %64s %s %n", name, type, proto, wait, user, path, &i);

		if (!strlen(name) || !strlen(type) || !strlen(proto) ||
		    !strlen(wait) || !strlen(user) || !strlen(path))
			die("Line %d: Not enough fields present", nline);

		s.name = strdup(name);

		if (!strcmp(type, "stream"))
			s.type = SOCK_STREAM;
		else if (!strcmp(type, "dgram"))
			s.type = SOCK_DGRAM;
		else if (!strcmp(type, "seqpacket"))
			s.type = SOCK_SEQPACKET;
		else
			die("%s (line %d): unknown type \"%s\"", name, nline, type);

		if (!strcmp(proto, "tcp") || !strcmp(proto, "tcp4")) {
			s.family = PF_INET;
			s.proto = IPPROTO_TCP;
			s.protostr = strdup("tcp");
		} else if (!strcmp(proto, "tcp6")) {
			s.family = PF_INET6;
			s.proto = IPPROTO_TCP;
			s.protostr = strdup("tcp");
		} else if (!strcmp(proto, "udp") || !strcmp(proto, "udp4")) {
			s.family = PF_INET;
			s.proto = IPPROTO_UDP;
			s.protostr = strdup("udp");
		} else if (!strcmp(proto, "udp6")) {
			s.family = PF_INET6;
			s.proto = IPPROTO_UDP;
			s.protostr = strdup("udp");
		} else if (!strcmp(proto, "unix")) {
			s.family = PF_UNIX;
			s.proto = 0;
			s.protostr = strdup("unix");
		} else {
			struct protoent *p = getprotobyname(proto);

			if (!p)
				die("%s (line %d): unknown protocol \"%s\"", name, nline, proto);

			s.family = PF_INET;
			s.proto = p->p_proto;
			s.protostr = strdup(p->p_name);
		}

		/* :<user>:<group>:<mode>:/path/to/sock */
		if (s.family == PF_UNIX) {
			int n = 0;

			for (c = s.name; *c; c++)
				if (*c == ':')
					n++;

			switch (n) {
			case 0:
				goto defunx;
			case 4:
				if (s.name[0] == ':')
					break;
			default:
				die("%s (line %d): bad syntax for path name", name, nline);
			}

			c = s.name + 1;

			x = strchr(c, ':');

			*x = 0;

			if (strlen(c)) {
				pw = getpwnam(c);

				if (!pw)
					die("%s (line %d): unknown user (owner) \"%s\"", name, nline, c);

				s.un.uid = pw->pw_uid;
			} else {
				s.un.uid = getuid();
			}

			c = x + 1;

			x = strchr(c, ':');

			*x = 0;

			if (strlen(c)) {
				gr = getgrnam(c);

				if (!gr)
					die("%s (line %d): unknown group (owner) \"%s\"", name, nline, c);

				s.un.gid = gr->gr_gid;
			} else {
				s.un.gid = getgid();
			}

			c = x + 1;

			x = strchr(c, ':');

			*x = 0;

			if (strlen(c))
				s.un.mode = (mode_t) strtol(c, NULL, 8);
			else
				s.un.mode = UDS_DEFAULT_MODE;

			c = strdup(x + 1);

			free(s.name);

			s.name = c;
		} else {
		defunx:
			s.un.uid = getuid();
			s.un.gid = getgid();
			s.un.mode = UDS_DEFAULT_MODE;
		}

		if (s.family != PF_UNIX) {
			port = getservbyname(s.name, s.protostr);

			if (!port)
				die("%s (line %d): unknown service for protocol \"%s\"", name, nline, s.protostr);
			s.port = ntohs(port->s_port);
		}

		if (!strcmp(wait, "wait")) {
			s.wait = 1;
			s.listenq = 1;
		} else if (!strcmp(wait, "nowait")) {
			s.wait = 0;
			s.listenq = LISTENQ;
		} else {
			die("%s (line %d): unknown wait field \"%s\"", name, nline, wait);
		}

		s.pid = 0;

		/* looks like we might have a group as well */
		if ((c = strchr(user, ':')))
			*c++ = 0;

		pw = getpwnam(user);

		if (!pw)
			die("%s (line %d): unknown user \"%s\"", name, nline, user);

		s.uid = pw->pw_uid;

		if (c && *c) {
			/* use group from group file */

			gr = getgrnam(c);

			if (!gr)
				die("%s (line %d): unknown group \"%s\"", name, nline, c);

			s.gid = gr->gr_gid;
		} else {
			/* use group from passwd file */

			s.gid = pw->pw_gid;
		}

		s.path = strdup(path);

		if (!strcmp(path, "internal")) {
			s.internal = 1;
			s.argv = NULL;
		} else {
			s.internal = 0;
			s.argv = makeargv(&line[i]);
		}

		verifyentry(&s, nline);

		s.aborted = 0;

		s.empty = 0;

		sservs = Realloc(sservs, sizeof(struct sserv) * (nservs + 1));

		sservs[nservs - 1] = s;

		nservs++;
	}

	fclose(fp);

	sservs[nservs - 1].name = NULL;

	return sservs;
}

