/*
 *
 * hal_vlynq.c
 * Description:
 * Basic VLYNQ driver implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/


/**************************************************************************/
/*      INCLUDES                                                          */
/**************************************************************************/
#include <linux/kernel.h>
#include <linux/init.h> 
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>

#include <puma5/puma5.h>
#include <puma5/puma5_hardware.h>
#include <generic/pal.h>

#include "_tistdtypes.h"
#include "hal_vlynq.h"

/**************************************************************************/
/*      DEFINES                                                           */
/**************************************************************************/

#define VLYNQ_RESET_SLAVE_GPIO    8

/* VLYNQ Registers */
#define VLYNQ_CONTROL_REGISTER    (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x04 )
#define VLYNQ_STATUS_REGISTER     (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x08 )
#define VLYNQ_TX_ADDRESS_MAP      (AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x1C )

#define VLYNQ_RX_ADDRESS_MAP_SIZE( index ) ((AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x20 ) + ( 0x8 * index ))
#define VLYNQ_RX_ADDRESS_MAP_OFFSET( index ) ((AVALANCHE_LOW_VLYNQ_CONTROL_BASE + 0x24 ) + ( 0x8 * index ))

/* Control register bits */
#define VLYNQ_CONTROL_CLOCK_DIRCTION_BIT    (1<<15)
#define VLYNQ_CONTROL_CLOCK_DIV_BITS( x )   (x<<16)
#define VLYNQ_CONTROL_RESET_BIT             (1<<0)

/* Status register bits */
#define VLYNQ_STATUS_LINK_BIT               (1<<0)

static struct proc_dir_entry *vlynq_proc = NULL;
static struct proc_dir_entry *vlynq_link_proc = NULL;

static int verbose = 1;

/* Static functions */
static void writeWord(unsigned int address, unsigned int value);
static unsigned int readWord(unsigned int address);
static int vlynq_read_link_proc(char *page, char **start, off_t off, int count, int *eof, void *data);
static int vlynq_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data);
static int vlynq_write_proc(struct file *file, const char *buffer, unsigned long count, void *data);

static irqreturn_t lnx_vlynq_root_isr(int irq, void * arg, struct pt_regs *regs);
static void VLYNQ_init(void);

/**************************************************************************/
/*      STATIC  Functions                                                 */
/**************************************************************************/

static void writeWord(unsigned int address, unsigned int value)
{
	volatile unsigned int *reg;
	reg = (unsigned int *) address;	
	if (verbose!=0)
		printk("Write: %p <-- %08x\n", reg, value);
	*reg = value;
	PAL_sysCacheFlush(PAL_OSMEM_ADDR_DAT, (void*)(reg), sizeof(unsigned int));
}

static unsigned int readWord(unsigned int address)
{
	volatile unsigned int *reg;
	unsigned int value;
	reg = (unsigned int *) address;	

	PAL_sysCacheInvalidate(PAL_OSMEM_ADDR_DAT, (void*)(reg), sizeof(unsigned int));

	if (verbose!=0)
		printk("Read: %p --> ", reg);
	value = *reg;
	if (verbose!=0)
		printk("%08x\n", value);
	return value;
}

static int vlynq_read_link_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    int len = 0;
    int link;
    int verbose_backup = verbose;
    
    verbose = 0;
    
    link = VLYNQ_get_link_status();
    
    verbose = verbose_backup;
    
    len += sprintf(page + len, "VLYNQ Link is %s.\n", link ? "UP" : "DOWN" );
    
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	
    return(len);
}

static int vlynq_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    int len = 0;
    
    len += sprintf(page + len, "Usage\n");
    len += sprintf(page + len, "\tv: verbose (is %s)\n", (verbose!=0)?"On":"Off");
    len += sprintf(page + len, "\tc:[l/r][i/o]: Setup clock direction: Local/Remote Input/Output\n");
    len += sprintf(page + len, "\td:[l/r] [0-7]: Setup clock divider: Local/Remote 0-7\n");
    len += sprintf(page + len, "\tt:[l/r] [address]: Setup TX map address: Local/Remote address\n");
    len += sprintf(page + len, "\tr:[l/r] [0-3] [address] [size]: Setup RX map address: Local/Remote index address size\n");
    len += sprintf(page + len, "\ts: Reset/Restore VLYNQ link\n");
    len += sprintf(page + len, "\tg:[l/r] Dump Local/Remote registers\n");
    len += sprintf(page + len, "\tR: Reset slave VLYNQ device\n");
    
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	
    return(len);
}


static int vlynq_write_proc(struct file *file,
                             const char *buffer,
                             unsigned long count, 
                             void *data)
{
        int len;
		char tmp_buffer[256];
		
        if(count > 256)
                len = 256;
        else
                len = count;

        if(copy_from_user(tmp_buffer, buffer, len))
                return -EFAULT;
        tmp_buffer[len-1] = '\0';

		switch (tmp_buffer[0])
		{
			case 'v':
				verbose = (verbose!=0)?0:1;
				if (verbose!=0)
					printk("verbose on\n");
				else
					printk("verbose off\n");
				break;
    		case 'c':
              {
                   vlynqType_e type;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     

                   printk("Setting %s clock to be ", type == VLYNQ_LOCAL ? "Local":"Remote");
                   if ( tmp_buffer[2]=='o' )
                   {
                        printk("source.\n");               
                        VLYNQ_setup_clock_direction( VLYNQ_SOURCE_CLOCK, type );
                   }
                   else
                   {
                        printk("sink.\n");
                        VLYNQ_setup_clock_direction( VLYNQ_SINK_CLOCK, type );
                   }
              }
              break;

    		case 'd':
              {
                   vlynqType_e type;
                   unsigned int div = 0;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     

                   sscanf( &tmp_buffer[2], "%d", &div );
                   
                   if( div > 7 )
                        div = 7;
                        
                   printk("Setting %s clock divider to be %d\n", type == VLYNQ_LOCAL ? "Local":"Remote", div);
                   VLYNQ_setup_clock_divider( (unsigned int)div, type );
              }
              break;

		     case 't':
              {
                   unsigned int addr;

                   vlynqType_e type;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     
                   
				printk("Set TX map pointer of %s vlynq address\n", type == VLYNQ_LOCAL ? "Local":"Remote");
				sscanf( &tmp_buffer[3], "%x", &addr );
				printk("Address = %p\n", (unsigned int *)addr);

                   VLYNQ_setup_tx_region( addr, type );
              }
              break;

		     case 'r':
              {
                   unsigned int addr;
                   unsigned int size;
                   unsigned int index;

                   vlynqType_e type;

                   if ( tmp_buffer[1]=='l' )
                   {
                        type = VLYNQ_LOCAL;
                   }
                   else
                   {    
                        type = VLYNQ_REMOTE;
                   }     
                   
				printk("Set RX map pointer of %s vlynq address\n", type == VLYNQ_LOCAL ? "Local":"Remote");
				sscanf( &tmp_buffer[3], "%d %x %x", &index, &addr, &size );
				printk("Index = %d, Address = %p, size = 0x%x\n", index, (unsigned int *)addr, size);

                   VLYNQ_setup_rx_region( addr, size, index, type );
              }
              break;
              
    		case 's':
                   printk( "%s link\n", VLYNQ_get_link_status() ? "Resetting":"Restoring" );
                   VLYNQ_reset_link( VLYNQ_get_link_status() );
                   break;
                   
    		case 'g':
              {
                   unsigned int i;
                   int last_verbose = verbose;
                   unsigned int base = AVALANCHE_LOW_VLYNQ_CONTROL_BASE;

                   if ( tmp_buffer[1]=='r' )
                        base += 0x80;
                   
                   verbose = 1;
                   for( i = 0; i < 0x20; i++ )
                        readWord( base + (i * 0x4) );
                   
                   verbose = last_verbose;     
              }
              break;

		     case 'R':
                   printk( "Resetting slave using GPIO #%d\n", VLYNQ_RESET_SLAVE_GPIO );
                   VLYNQ_reset_slave();
                   break;
         
              default:
                   printk( "Unsupported command.\n");
                   break;
		}
		
        return len;
}

/* Wrapper for vlynq ISR */
static irqreturn_t lnx_vlynq_root_isr(int irq, void * arg, struct pt_regs *regs)
{   
   return IRQ_HANDLED;
}

static void VLYNQ_init(void)
{
    int irq;
    unsigned int i;

	/* Take out of reset VLYNQ and GPIO PSC */
    PAL_sysPscSetModuleState( PSC_GPIO, PSC_ENABLE );		
    PAL_sysPscSetModuleState( PSC_VLYNQ, PSC_ENABLE );	
    PAL_sysPscSetDomainState( PSC_PD_ARM, PSC_ENABLE );

	/* Disable aux GPIO */
    //PAL_sysGpioInit();

    for (i=10+32; i<=18+32; i++)
    {
		PAL_sysGpioCtrl(i, FUNCTIONAL_PIN, GPIO_OUTPUT_PIN);
    }
    
    PAL_sysGpioCtrl(VLYNQ_RESET_SLAVE_GPIO, GPIO_PIN, GPIO_OUTPUT_PIN);     
    PAL_sysGpioOutBit(VLYNQ_RESET_SLAVE_GPIO, 1 );
	
    /* Request IRQ */
    irq = LNXINTNUM(AVALANCHE_LOW_VLYNQ_INT);
    request_irq(irq, lnx_vlynq_root_isr, 0, "vlynq", NULL);
	 printk("VLYNQ driver initialization complete.\n");
}

/**************************************************************************/
/*      INTERFACE  Functions                                              */
/**************************************************************************/

int VLYNQ_reset_slave( void )
{
    PAL_sysGpioOutBit( VLYNQ_RESET_SLAVE_GPIO, 0 );
    msleep( 200 );    
    PAL_sysGpioOutBit( VLYNQ_RESET_SLAVE_GPIO, 1 );
    
    return 0;
}

/* Reset VLYNQ link: reset = 1 reset, reset = 0 out of reset */
int VLYNQ_reset_link( unsigned int reset )
{
    unsigned int reg;

    reg = readWord( VLYNQ_CONTROL_REGISTER );
    
    if( reset )
        reg |= VLYNQ_CONTROL_RESET_BIT;
    else
        reg &= ~VLYNQ_CONTROL_RESET_BIT;
    
    writeWord( VLYNQ_CONTROL_REGISTER, reg );
    
    return 0;
}

/* Get VLYNQ link status: 1 Link up, 0 link down */
int VLYNQ_get_link_status( void )
{
    unsigned int reg;

    reg = readWord( VLYNQ_STATUS_REGISTER );
    
    return reg & VLYNQ_STATUS_LINK_BIT ? 1 : 0;
}

/* Setup clock divider: 0 to 7 (+1) */
int VLYNQ_setup_clock_divider( unsigned int div, vlynqType_e type )
{
    unsigned int reg;

    if( div > 7 )
         return -1;

    reg = readWord( VLYNQ_CONTROL_REGISTER + type );
    
    reg &= ~VLYNQ_CONTROL_CLOCK_DIV_BITS(7);
    reg |= VLYNQ_CONTROL_CLOCK_DIV_BITS( div );
    
    writeWord( VLYNQ_CONTROL_REGISTER + type, reg );

    return 0;
}

/* Setup clock direction: Source (output) or Sink (input) */
int VLYNQ_setup_clock_direction( vlynqClock_e clock, vlynqType_e type )
{
    unsigned int reg;

    if ( type == VLYNQ_LOCAL )
    {
         if ( clock ==  VLYNQ_SOURCE_CLOCK )
         {
              /* Setup BOOTCR */
              reg = readWord( AVALANCHE_DCL_BOOTCR );
              reg |= (1<<PAL_SYS_CLKC_VLYNQ_DIR_BIT);
              writeWord( AVALANCHE_DCL_BOOTCR, reg );
         }
         else
         {
              /* Setup BOOTCR */
              reg = readWord( AVALANCHE_DCL_BOOTCR );
              reg &= ~(1<<PAL_SYS_CLKC_VLYNQ_DIR_BIT);
              writeWord( AVALANCHE_DCL_BOOTCR, reg );
         }
    }
    
    reg = readWord( VLYNQ_CONTROL_REGISTER + type );
    
    if( clock == VLYNQ_SOURCE_CLOCK )
        reg |= VLYNQ_CONTROL_CLOCK_DIRCTION_BIT;
    else
        reg &= ~VLYNQ_CONTROL_CLOCK_DIRCTION_BIT;
    
    writeWord( VLYNQ_CONTROL_REGISTER + type, reg );

    return 0;
}

/* TX region: The address which will be used for VLYNQ access */
int VLYNQ_setup_tx_region( unsigned int address, vlynqType_e type )
{
    writeWord( VLYNQ_TX_ADDRESS_MAP + type, address );

    return 0;
}

/* RX regions: The address windows which will be available to VLYNQ */
int VLYNQ_setup_rx_region( unsigned int address, unsigned int size, unsigned int index, vlynqType_e type )
{
    if( index > 3 )
         return -1;
         
    writeWord( VLYNQ_RX_ADDRESS_MAP_SIZE( index ) + type, size );
    writeWord( VLYNQ_RX_ADDRESS_MAP_OFFSET( index ) + type, address );
    
    return 0;
}

/* initialize module (and interrupt) */
static int __init vlynqInit (void) 
{
    vlynq_proc = create_proc_entry("vlynq", 0, NULL);
    vlynq_link_proc = create_proc_read_entry("vlynq_link", 0, NULL, vlynq_read_link_proc, NULL);    
	 vlynq_proc->read_proc = vlynq_read_proc;
	 vlynq_proc->write_proc = vlynq_write_proc;

	 VLYNQ_init();
    
	 return 0;
}

/* close and cleanup module */
static void __exit vlynqCleanup (void) 
{
	remove_proc_entry("vlynq", NULL);
	remove_proc_entry("vlynq_link", NULL);   
}

module_init(vlynqInit);
module_exit(vlynqCleanup);
MODULE_AUTHOR("Texas Instruments, Inc");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Basic VLYNQ Interface driver");

EXPORT_SYMBOL( VLYNQ_reset_link );
EXPORT_SYMBOL( VLYNQ_get_link_status );
EXPORT_SYMBOL( VLYNQ_setup_clock_divider );
EXPORT_SYMBOL( VLYNQ_setup_clock_direction );
EXPORT_SYMBOL( VLYNQ_setup_tx_region );
EXPORT_SYMBOL( VLYNQ_setup_rx_region );
EXPORT_SYMBOL( VLYNQ_reset_slave );
